(define-constant architecture 'x86_64)
(define-constant ptr-bits 64)

(define-constant asm-arg-reg-max 5)
(define-constant asm-arg-reg-cnt 3)
(define-constant asm-fpreg-max 8)

(define-constant native-endianness 'little)

(define-constant unaligned-floats #t)
(define-constant unaligned-integers #t)
(define-constant integer-divide-instruction #t)
(define-constant popcount-instruction #t)
