;; This template is turned into a machine-specific ".def" file
;; by the `workarea` script

(define-constant machine-type (constant machine-type-$(M)))
(features iconv expeditor pbchunk)

(define-constant-default native-endianness '$(Mend))

(include "pbcommon$(Mword).def")
(include "pbcommon.def")
(include "default.def")
