(define-constant architecture 'pb)

(define-constant typedef-i8 "int8_t")
(define-constant typedef-u8 "uint8_t")
(define-constant typedef-i16 "int16_t")
(define-constant typedef-u16 "uint16_t")
(define-constant typedef-i32 "int32_t")
(define-constant typedef-u32 "uint32_t")
(define-constant typedef-i64 "int64_t")
(define-constant typedef-u64 "uint64_t")
(define-constant typedef-string-char "uint32_t")

(define-constant asm-arg-reg-max 7)
(define-constant asm-arg-reg-cnt 3)
(define-constant asm-fpreg-max 8)

(define-constant max-float-alignment 8)
(define-constant max-integer-alignment 8)

(define-constant-default native-endianness 'unknown)

(define-constant unaligned-floats #f)
(define-constant unaligned-integers #f)

(define-constant integer-divide-instruction #t)

(include "default.def")
