;; This template is turned into a machine-specific ".def" file
;; by the `machine.def` script

(define-constant machine-type (constant machine-type-$(M)))
(features iconv expeditor pthreads)
(include "$(March).def")
(include "default.def")
